// BitmapItem.cpp: implementation of the CBitmapItem class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "BitmapItem.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

void TransBlt (HDC hDstDC, int x, int y, int cx, int cy, HDC hSrcDC)
{
	//     (   )
	HDC hSrcCopyDC = CreateCompatibleDC (hSrcDC);
	HBITMAP hSrcCopyBmp = CreateCompatibleBitmap (hSrcDC, cx, cy);
	HBITMAP hOldSrcCopyBmp = (HBITMAP) SelectObject (hSrcCopyDC, hSrcCopyBmp);
	BitBlt (hSrcCopyDC, 0, 0, cx, cy, hSrcDC, 0, 0, SRCCOPY);

	//   ( ) =     0, 0
	COLORREF OldColor = SetBkColor (hSrcDC, GetPixel (hSrcDC, 0, 0));

	//   ,   0 - , 1 - 
	HDC hMaskDC = CreateCompatibleDC (hSrcDC);
	HBITMAP hMaskBmp = CreateCompatibleBitmap (hMaskDC, cx, cy);
	HBITMAP hOldMaskBmp = (HBITMAP) SelectObject (hMaskDC, hMaskBmp);	
	BitBlt (hMaskDC, 0, 0, cx, cy, hSrcDC, 0, 0, SRCCOPY);
	
	SetBkColor (hSrcDC, OldColor);
	
	//    
	HDC hInvMaskDC = CreateCompatibleDC (hSrcDC);
	HBITMAP hInvMaskBmp = CreateCompatibleBitmap (hInvMaskDC, cx, cy);
	HBITMAP hOldInvMaskBmp = (HBITMAP) SelectObject (hInvMaskDC, hInvMaskBmp);	
	BitBlt (hInvMaskDC, 0, 0, cx, cy, hMaskDC, 0, 0, NOTSRCCOPY);

	//    
	BitBlt (hDstDC, x, y, cx, cy, hMaskDC, 0, 0, SRCAND);
	BitBlt (hSrcCopyDC, 0, 0, cx, cy, hInvMaskDC, 0, 0, SRCAND);
	BitBlt (hDstDC, x, y, cx, cy, hSrcCopyDC, 0, 0, SRCPAINT);

	// 
	SelectObject (hSrcCopyDC, hOldSrcCopyBmp);
	DeleteObject (hSrcCopyBmp);
	DeleteDC (hSrcCopyDC);

	SelectObject (hMaskDC, hOldMaskBmp);
	DeleteObject (hMaskBmp);
	DeleteDC (hMaskDC);

	SelectObject (hInvMaskDC, hOldInvMaskBmp);
	DeleteObject (hInvMaskBmp);
	DeleteDC (hInvMaskDC);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CBitmapItem::CBitmapItem ()
{
	m_hBmp = NULL;
}

CBitmapItem::~CBitmapItem()
{
	if (m_hBmp != NULL) {
		DeleteObject (m_hBmp);
		m_hBmp = NULL;
	}
}

void CBitmapItem::Create (CMenu* menu, UINT nID, UINT nBmpID, UINT cx, UINT cy,
						  LPCTSTR szLabel, UINT nFlags, BOOL bChecked)
{
	ASSERT (menu);

	if (m_hBmp != NULL) {
		DeleteObject (m_hBmp);
		m_hBmp = NULL;
	}

	//   
	CString strLabel;
	if (szLabel == NULL)
		menu->GetMenuString (nID, strLabel, nFlags);
	else
		strLabel = szLabel;

	HDC hScreenDC = GetDC (NULL);
	ASSERT (hScreenDC);

	//  
	HBITMAP hSrcBmp = LoadBitmap (AfxGetInstanceHandle (), MAKEINTRESOURCE (nBmpID));
	ASSERT (hSrcBmp);
	HDC hSrcDC = CreateCompatibleDC (hScreenDC);
	ASSERT (hSrcDC);
	BITMAP bm;
	GetObject (hSrcBmp, sizeof (BITMAP), &bm);
	HBITMAP hSrcOldBmp = (HBITMAP) SelectObject (hSrcDC, hSrcBmp);

	//  
	int width = cx;
	int height = (cy == 0) ? bm.bmHeight + 4 : cy;
	HDC hDstDC = CreateCompatibleDC (hScreenDC);
	ASSERT (hDstDC);
	m_hBmp = CreateCompatibleBitmap (hScreenDC, width, height);
	ASSERT (m_hBmp);
	HBITMAP hDstOldBmp = (HBITMAP) SelectObject (hDstDC, m_hBmp);

	// 
	CRect rc (0, 0, width, height);
	HBRUSH hBrush = CreateSolidBrush (GetSysColor (COLOR_MENU));
	ASSERT (hBrush);
	FillRect (hDstDC, &rc, hBrush);
	DeleteObject (hBrush);

	TransBlt (hDstDC, 2, 2, bm.bmWidth, bm.bmHeight, hSrcDC);
	
	SetBkMode (hDstDC, TRANSPARENT);

	HFONT hFont = CreateFont (- (height / 3) * 2, 0, 0, 0,
		((GetMenuDefaultItem (menu->GetSafeHmenu (), nFlags & MF_BYPOSITION,
			GMDI_USEDISABLED) == nID) ? FW_SEMIBOLD : 0),
		FALSE, FALSE, FALSE, RUSSIAN_CHARSET, OUT_TT_ONLY_PRECIS,
		CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, VARIABLE_PITCH | FF_ROMAN, _T("Arial"));
	ASSERT (hFont);
	HFONT hOldFont = (HFONT) SelectObject (hDstDC, hFont);

	// 
	SetTextColor (hDstDC, GetSysColor (COLOR_MENUTEXT));
	CRect textrc (bm.bmWidth + 7, 0, width - 1, height - 1);
	DrawText (hDstDC, strLabel, strLabel.GetLength (), textrc,
		DT_LEFT | DT_SINGLELINE | DT_VCENTER | DT_EXPANDTABS);
	
	// 
	SelectObject (hSrcDC, hSrcOldBmp);
	DeleteDC (hSrcDC);
	DeleteObject (hSrcBmp);

	SelectObject (hDstDC, hOldFont);
	DeleteObject (hFont);
	SelectObject (hDstDC, hDstOldBmp);
	DeleteDC (hDstDC);

	ReleaseDC (NULL, hScreenDC);

	//    
	menu->ModifyMenu (nID, MF_BITMAP | nFlags | (bChecked ? MF_CHECKED : MF_UNCHECKED),
		nID, (LPCTSTR) m_hBmp);
}
