// DXImage.cpp: implementation of the CDCDXImage class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DXMisc.h"
#include "DXImage.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC (CDXImage, CObject);

CDXImage::CDXImage()
{
	::InitializeCriticalSection (&m_cs);
	m_lpDDSurface3 = NULL;
	bm.bmWidth = 0;
	bm.bmHeight = 0;
	m_LoadType = 3;
}

CDXImage::~CDXImage()
{
	Destroy ();
	::DeleteCriticalSection (&m_cs);
}

BOOL CDXImage::LoadFile (LPDIRECTDRAW2 lpDD, LPCTSTR szBitmap)
{
	::EnterCriticalSection (&m_cs);
	BOOL ret = LoadHelper (lpDD, szBitmap, FALSE);
	::LeaveCriticalSection (&m_cs);
	return ret;
}

BOOL CDXImage::LoadResource (LPDIRECTDRAW2 lpDD, LPCTSTR szBitmap)
{
	::EnterCriticalSection (&m_cs);
	BOOL ret = LoadHelper (lpDD, szBitmap, TRUE);
	::LeaveCriticalSection (&m_cs);
	return ret;
}

BOOL CDXImage::Reload (void)
{
	switch (m_LoadType) {
		case 0:
			return LoadHelper (m_lpDD, m_szBitmap, FALSE);
		case 1:
			return LoadHelper (m_lpDD, m_szBitmap, TRUE);
	}	
	return Create (m_lpDD, bm.bmWidth, bm.bmHeight);
}

BOOL CDXImage::LoadHelper (LPDIRECTDRAW2 lpDD, LPCTSTR szBitmap, BOOL bResource)
{
	HBITMAP hbm = NULL;

	DestroyHelper ();
	if (lpDD == NULL || szBitmap == NULL)
		return FALSE;

	m_lpDD = lpDD;
	m_szBitmap = szBitmap;
	m_LoadType = (bResource ? 1 : 0);

	if (bResource) {
		hbm = (HBITMAP) LoadImage (GetModuleHandle (NULL),
			szBitmap, IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
		if (hbm == NULL) {
			TRACE (_T("CDXImage::LoadHelper:LoadImage resource \"%s\"\n"), szBitmap);
			::LeaveCriticalSection (&m_cs);
			return FALSE;
		}
	} else {
		hbm = (HBITMAP) LoadImage (NULL,
			szBitmap, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);
		if (hbm == NULL) {
			TRACE (_T("CDXImage::LoadHelper:LoadImage file \"%s\"\n"), szBitmap);
			return FALSE;
		}
	}

    BOOL ret = LoadFromHandle (lpDD, hbm);
    DeleteObject (hbm);
	return ret;
}

BOOL CDXImage::LoadFromHandle (LPDIRECTDRAW2 lpDD, HBITMAP hbm)
{
	if (hbm == NULL)
		return FALSE;

	//  select bitmap into a memoryDC so we can use it.
	HDC hdcImage = CreateCompatibleDC (NULL);
	if (!hdcImage)
		return FALSE;

	// get size of the bitmap
    GetObject (hbm, sizeof(bm), &bm);

	SelectObject (hdcImage, hbm);

    // create surface.
	LPDIRECTDRAWSURFACE lpDDSurface;
	DDSURFACEDESC ddsd;
	memset (&ddsd, 0, sizeof (ddsd));
	ddsd.dwSize = sizeof (ddsd);
    ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_CAPS;
	ddsd.dwWidth = bm.bmWidth;
	ddsd.dwHeight = bm.bmHeight;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	HRESULT ddrval = lpDD->CreateSurface (&ddsd, &lpDDSurface, NULL);
	if (ddrval != DD_OK || lpDDSurface == NULL) {
		debugDXReportError (_T ("CDXImage::LoadFromHandle:CreateSurface"), ddrval);
		DeleteDC (hdcImage);
		DestroyHelper ();
		return FALSE;
	}

	ddrval = lpDDSurface->QueryInterface (IID_IDirectDrawSurface3,
		(LPVOID*) &m_lpDDSurface3);
	lpDDSurface->Release ();
	lpDDSurface = NULL;
	if (ddrval != DD_OK || m_lpDDSurface3 == NULL) {
		debugDXReportError (_T ("CDXImage::LoadFromHandle:QueryInterface"), ddrval);
		DeleteDC (hdcImage);
		DestroyHelper ();
		return FALSE;
	}

	//   ( GDI)
	HDC hdc = NULL;
	ddrval = m_lpDDSurface3->GetDC (&hdc);
	if (ddrval != DD_OK && hdc == NULL) {
		debugDXReportError (_T ("CDXImage::LoadFromHandle:GetDC"), ddrval);
		DeleteDC (hdcImage);
		DestroyHelper ();
		return FALSE;
	}
	BitBlt (hdc, 0, 0, ddsd.dwWidth, ddsd.dwHeight, hdcImage, 0, 0, SRCCOPY);
	m_lpDDSurface3->ReleaseDC (hdc);
		
	//   -    0,0
	SetColorKey ();

    DeleteDC (hdcImage);
    return TRUE;
}

BOOL CDXImage::Create (LPDIRECTDRAW2 lpDD, UINT cx, UINT cy)
{
	::EnterCriticalSection (&m_cs);
	
	DestroyHelper ();

	if (lpDD == NULL || cx == 0 || cy == 0) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	m_lpDD = lpDD;
	m_szBitmap = NULL;
	m_LoadType = 3;

    // create surface.
	LPDIRECTDRAWSURFACE lpDDSurface;
	DDSURFACEDESC ddsd;
	memset (&ddsd, 0, sizeof (ddsd));
	ddsd.dwSize = sizeof (ddsd);
    ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_CAPS;
	ddsd.dwWidth = bm.bmWidth = cx;
	ddsd.dwHeight = bm.bmHeight = cy;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	HRESULT ddrval = lpDD->CreateSurface (&ddsd, &lpDDSurface, NULL);
	if (ddrval != DD_OK || lpDDSurface == NULL) {
		debugDXReportError (_T ("CDXImage::Create:CreateSurface"), ddrval);
		DestroyHelper ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	ddrval = lpDDSurface->QueryInterface (IID_IDirectDrawSurface3,
		(LPVOID*) &m_lpDDSurface3);
	lpDDSurface->Release ();
	lpDDSurface = NULL;
	if (ddrval != DD_OK || m_lpDDSurface3 == NULL) {
		debugDXReportError (_T ("CDXImage::Create:QueryInterface"), ddrval);
		DestroyHelper ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}
	
BOOL CDXImage::Fill (COLORREF color)
{
	::EnterCriticalSection (&m_cs);

	if (m_lpDDSurface3 == NULL) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	//   
	DDBLTFX ddbltfx;
	ddbltfx.dwSize = sizeof (ddbltfx);
	ddbltfx.dwFillColor = color;

	BOOL result = TRUE;
	while (1) {
		//   : 1) Restore, 2) Reload
		if (m_lpDDSurface3->IsLost () == DDERR_SURFACELOST)
			if (m_lpDDSurface3->Restore () != DD_OK)
				if (!Reload ()) {
					result = FALSE;
					break;
				}

		HRESULT ddrval = m_lpDDSurface3->Blt (NULL, NULL, NULL, DDBLT_COLORFILL, &ddbltfx);
		if (ddrval == DD_OK)
			//  
			break;
		else
			if (ddrval != DDERR_WASSTILLDRAWING) {
				//    ,    
				debugDXReportError (_T ("CDXImage::Fill:Blt"), ddrval);
				result = FALSE;
				break;
			}
	}
	
	//   -    0,0
	SetColorKey ();

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

BOOL CDXImage::FastDraw (LPDIRECTDRAWSURFACE3 lpDDDestSurface, CRect rect, int x, int y,
						 DWORD dwFlags)
{
	::EnterCriticalSection (&m_cs);

	if (lpDDDestSurface == NULL || m_lpDDSurface3 == NULL) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	BOOL result = TRUE;
	while (1) {
		//   : 1) Restore, 2) Reload
		if (m_lpDDSurface3->IsLost () == DDERR_SURFACELOST)
			if (m_lpDDSurface3->Restore () != DD_OK)
				if (!Reload ()) {
					result = FALSE;
					break;
				}

		HRESULT ddrval = lpDDDestSurface->BltFast (x, y, m_lpDDSurface3, rect, dwFlags);
		if (ddrval == DD_OK)
			//  
			break;
		else
			if (ddrval == DDERR_SURFACELOST) {
				//   
				ddrval = lpDDDestSurface->Restore ();
				if (ddrval != DD_OK) {
					//   
					debugDXReportError (_T ("CDXImage::FastDraw:Restore"), ddrval);
					result = FALSE;
					break;
				}
			}
			else
				if (ddrval != DDERR_WASSTILLDRAWING) {
					//    ,    
					debugDXReportError (_T ("CDXImage::FastDraw:BltFast"), ddrval);
					result = FALSE;
					break;
				}
	}

	::LeaveCriticalSection (&m_cs);

	return result;
}

BOOL CDXImage::Draw (LPDIRECTDRAWSURFACE3 lpDDDestSurface, CRect rect, int x, int y,
							  DWORD dwFlags, LPDDBLTFX lpDDBltFx)
{
	::EnterCriticalSection (&m_cs);

	if (lpDDDestSurface == NULL || m_lpDDSurface3 == NULL) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	CRect rc (x, y, x + rect.Width () - 1, y + rect.Height () - 1);
	BOOL result = TRUE;
	while (1) {
		//   : 1) Restore, 2) Reload
		if (m_lpDDSurface3->IsLost () == DDERR_SURFACELOST)
			if (m_lpDDSurface3->Restore () != DD_OK)
				if (!Reload ()) {
					result = FALSE;
					break;
				}

		HRESULT ddrval = lpDDDestSurface->Blt (rc, m_lpDDSurface3, rect, dwFlags, lpDDBltFx);
		if (ddrval == DD_OK)
			//  
			break;
		else
			if (ddrval == DDERR_SURFACELOST) {
				//   
				ddrval = lpDDDestSurface->Restore ();
				if (ddrval != DD_OK) {
					//   
					debugDXReportError (_T ("CDXImage::Draw:Restore"), ddrval);
					result = FALSE;
					break;
				}
			}
			else
				if (ddrval != DDERR_WASSTILLDRAWING) {
					//    ,    
					debugDXReportError (_T ("CDXImage::Draw:Blt"), ddrval);
					result = FALSE;
					break;
				}
	}

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

void CDXImage::Destroy (void)
{
	::EnterCriticalSection (&m_cs);
	DestroyHelper ();
	::LeaveCriticalSection (&m_cs);
}

void CDXImage::DestroyHelper (void)
{
	if (m_lpDDSurface3 != NULL) {
		m_lpDDSurface3->Release ();
		m_lpDDSurface3 = NULL;
	}
}

BOOL CDXImage::SetColorKey (COLORREF rgb)
{
	if (m_lpDDSurface3 == NULL)
		return FALSE;

	DDCOLORKEY ddck;
    ddck.dwColorSpaceLowValue  = ColorMatch (rgb);
    ddck.dwColorSpaceHighValue = ddck.dwColorSpaceLowValue;
	
	if (ddck.dwColorSpaceLowValue == CLR_INVALID)
		return FALSE;

    return (m_lpDDSurface3->SetColorKey (DDCKEY_SRCBLT, &ddck) == DD_OK);
}

DWORD CDXImage::ColorMatch (COLORREF rgb)
{
	if (m_lpDDSurface3 == NULL)
		return CLR_INVALID;

	COLORREF rgbT;
    HDC hdc;
    DWORD dw = CLR_INVALID;
    HRESULT hres;

    //  use GDI SetPixel to color match for us
    if (rgb != CLR_INVALID && m_lpDDSurface3->GetDC (&hdc) == DD_OK) {
		rgbT = GetPixel (hdc, 0, 0);             // save current pixel value
		SetPixel (hdc, 0, 0, rgb);               // set our value
		m_lpDDSurface3->ReleaseDC (hdc);
    }

    // now lock the surface so we can read back the converted color
    DDSURFACEDESC ddsd;
	ddsd.dwSize = sizeof(ddsd);
	while ((hres = m_lpDDSurface3->Lock (NULL, &ddsd, 0, NULL)) ==
		DDERR_WASSTILLDRAWING);

    if (hres == DD_OK) {
		dw = *(DWORD *)ddsd.lpSurface;	// get DWORD
		if (ddsd.ddpfPixelFormat.dwRGBBitCount < 32)
			dw &= (1 << ddsd.ddpfPixelFormat.dwRGBBitCount) - 1;  // mask it to bpp
		m_lpDDSurface3->Unlock (NULL);
    }

    //  now put the color that was there back.
	if (rgb != CLR_INVALID && m_lpDDSurface3->GetDC (&hdc) == DD_OK) {
		SetPixel (hdc, 0, 0, rgbT);
		m_lpDDSurface3->ReleaseDC (hdc);
	}

    return dw;
}

HDC CDXImage::GetDC (void)
{
	::EnterCriticalSection (&m_cs);
	if (m_lpDDSurface3 == NULL) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}
	HDC hdc = NULL;
	if (m_lpDDSurface3->GetDC (&hdc) == DD_OK) {
		::LeaveCriticalSection (&m_cs);
		return hdc;
	} else {
		::LeaveCriticalSection (&m_cs);
		return NULL;
	}
}

BOOL CDXImage::ReleaseDC (HDC hdc)
{
	::EnterCriticalSection (&m_cs);
	if (hdc == NULL || m_lpDDSurface3 == NULL) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}
	m_lpDDSurface3->ReleaseDC (hdc);
	::LeaveCriticalSection (&m_cs);
	return TRUE;
}
