// DXScreen.cpp: implementation of the CDCDXScreen class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DXMisc.h"
#include "DXScreen.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNAMIC (CDXScreen, CObject);

CDXScreen::CDXScreen ()
{
	::InitializeCriticalSection (&m_cs);

	m_lpDD2 = NULL;
	m_lpDDSPrimary3 = NULL;
	m_lpDDSBack3 = NULL;
	m_lpDDClipper = NULL;
	m_pWnd = NULL;
	m_cx = m_cy = 0;
}

CDXScreen::~CDXScreen ()
{
	Destroy ();

	::DeleteCriticalSection (&m_cs);
}

BOOL CDXScreen::Create (CWnd* pWnd, UINT cx, UINT cy)
{
	::EnterCriticalSection (&m_cs);

	//   
	Destroy ();
	
	if (pWnd == NULL)
		m_pWnd = AfxGetMainWnd ();
	else
		m_pWnd = pWnd;

	m_cx = cx;
	m_cy = cy;

	//   
	GetWindowRect (m_pWnd->GetSafeHwnd (), &m_Rect);
	if (cx != 0 && cy != 0) {
		m_Rect.right = m_Rect.left + cx - 1;
		m_Rect.bottom = m_Rect.top + cy - 1;
	}

	//   DirectDraw
	LPDIRECTDRAW lpDD;
	HRESULT ddrval = DirectDrawCreate (NULL, &lpDD, NULL);
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:DirectDrawCreate"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}
	ddrval = lpDD->QueryInterface (IID_IDirectDraw2, (LPVOID*) &m_lpDD2);
	lpDD->Release ();
	lpDD = NULL;
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:QueryInterface (DirectDraw)"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	//  
	ddrval = m_lpDD2->SetCooperativeLevel (m_pWnd->GetSafeHwnd (), DDSCL_NORMAL);
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:SetCooperativeLevel"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	::LeaveCriticalSection (&m_cs);

	return CreateSurfaces ();
}

BOOL CDXScreen::CreateSurfaces (void)
{
	::EnterCriticalSection (&m_cs);

	DestroySurfaces ();

	DDSURFACEDESC ddsd;
	memset (&ddsd, 0, sizeof (ddsd));
	ddsd.dwSize = sizeof (ddsd);
	
	//   
	LPDIRECTDRAWSURFACE lpDDSPrimary;
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	
	HRESULT ddrval = m_lpDD2->CreateSurface (&ddsd, &lpDDSPrimary, NULL);
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:CreateSurface (Primary)"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}
	ddrval = lpDDSPrimary->QueryInterface (IID_IDirectDrawSurface3,
		(LPVOID*) &m_lpDDSPrimary3);
	lpDDSPrimary->Release ();
	lpDDSPrimary = NULL;
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:QueryInterface (Primary)"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	//   
	LPDIRECTDRAWSURFACE lpDDSBack;
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = m_Rect.right - m_Rect.left;
	ddsd.dwHeight = m_Rect.bottom - m_Rect.top;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddrval = m_lpDD2->CreateSurface (&ddsd, &lpDDSBack, NULL);
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:CreateSurface (Secondary)"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}
	ddrval = lpDDSBack->QueryInterface (IID_IDirectDrawSurface3, (LPVOID*) &m_lpDDSBack3);
	lpDDSBack->Release ();
	lpDDSBack = NULL;
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:QueryInterface (Secondary)"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	//       
	ddrval = m_lpDD2->CreateClipper (0, &m_lpDDClipper, NULL);
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:CreateClipper"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}
	ddrval = m_lpDDClipper->SetHWnd (0, m_pWnd->GetSafeHwnd ());
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:SetHWnd"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}
	m_lpDDSPrimary3->SetClipper (m_lpDDClipper);
	if (ddrval != DD_OK) {
		debugDXReportError (_T ("CDXScreen::Create:SetClipper"), ddrval);
		Destroy ();
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	::LeaveCriticalSection (&m_cs);
	return TRUE;
}

void CDXScreen::Destroy (void)
{
	DestroySurfaces ();

	//  DirectDraw
    if (m_lpDD2 != NULL) {
		m_lpDD2->Release();
		m_lpDD2 = NULL;
    }
}

void CDXScreen::DestroySurfaces (void)
{
	//  
	if(m_lpDDClipper != NULL) {
		m_lpDDClipper->Release ();
		m_lpDDClipper = NULL;
	}
	if(m_lpDDSBack3 != NULL) {
		m_lpDDSBack3->Release ();
		m_lpDDSBack3 = NULL;
	}
	if(m_lpDDSPrimary3 != NULL) {
		m_lpDDSPrimary3->Release ();
		m_lpDDSPrimary3 = NULL;
	}
}

HDC CDXScreen::GetPrimaryDC (void)
{
	if (m_lpDDSPrimary3 == NULL)
		return FALSE;

	HDC hdc = NULL;
	if (m_lpDDSPrimary3->GetDC (&hdc) == DD_OK)
		return hdc;
	else
		return NULL;
}

BOOL CDXScreen::ReleasePrimaryDC (HDC hdc)
{
	if (hdc == NULL || m_lpDDSPrimary3 == NULL)
		return FALSE;

	m_lpDDSPrimary3->ReleaseDC (hdc);
	return TRUE;
}

HDC CDXScreen::GetBackDC (void)
{
	if (m_lpDDSBack3 == NULL)
		return FALSE;

	HDC hdc = NULL;
	if (m_lpDDSBack3->GetDC (&hdc) == DD_OK)
		return hdc;
	else
		return NULL;
}

BOOL CDXScreen::ReleaseBackDC (HDC hdc)
{
	if (hdc == NULL || m_lpDDSBack3 == NULL)
		return FALSE;

	m_lpDDSBack3->ReleaseDC (hdc);
	return TRUE;
}

BOOL CDXScreen::Draw (void)
{
	::EnterCriticalSection (&m_cs);

	if (m_lpDDSPrimary3 == NULL || m_lpDDSBack3 == NULL) {
		::LeaveCriticalSection (&m_cs);
		return FALSE;
	}

	//    ,     
	RECT dstRect, srcRect;
	srcRect.left = 0;
	srcRect.top = 0;
	GetWindowRect (m_pWnd->GetSafeHwnd (), &dstRect);
	dstRect.right = dstRect.left + (m_Rect.right - m_Rect.left) - 1;
	srcRect.right = (m_Rect.right - m_Rect.left) - 1;
	dstRect.bottom = dstRect.top + (m_Rect.bottom - m_Rect.top) - 1;
	srcRect.bottom = (m_Rect.bottom - m_Rect.top) - 1;

	while (1) {
		HRESULT ddrval = m_lpDDSPrimary3->Blt (&dstRect, m_lpDDSBack3,
			&srcRect, DDBLT_WAIT, NULL);
		if (ddrval == DD_OK)
			//  
			break;
		else
			if (ddrval == DDERR_SURFACELOST) {
				//   
				ddrval = m_lpDDSPrimary3->Restore ();
				if (ddrval != DD_OK) {
					//    
					::LeaveCriticalSection (&m_cs);
					return CreateSurfaces ();
				}
			}
			else
				if (ddrval != DDERR_WASSTILLDRAWING) {
					//    ,    
					debugDXReportError (_T ("CDXScreen::Flip:Blt"), ddrval);
					::LeaveCriticalSection (&m_cs);
					return FALSE;
				}
	}
	::LeaveCriticalSection (&m_cs);
	return TRUE;
}
